#!/usr/bin/env python
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt

from gym.envs.classic_control import CartPoleEnv

__author__ = "Anand Subramoney"

env = CartPoleEnv()
n_actions = env.action_space.n
n_obs = env.observation_space.high.__len__()

print_per_episode = 100

max_episode_len = 200
batch_size = 200
n_train_trials = 1000
n_test_trials = 100
gamma = 0.9999
learning_rate = 0.01

constant_baseline = 1.

state_holder = tf.placeholder(dtype=tf.float32, shape=(None, n_obs), name='symbolic_state')
actions_one_hot_holder = tf.placeholder(dtype=tf.float32, shape=(None, n_actions),
                                        name='symbolic_actions_one_hot_holder')

discounted_rewards_holder = tf.placeholder(dtype=tf.float32, shape=None, name='symbolic_reward')

with tf.name_scope('linear'):
    theta_var = tf.Variable(initial_value=np.random.rand(n_obs, n_actions), validate_shape=True,
                            trainable=True, dtype=tf.float32, name='theta')
    bias_var = tf.Variable(initial_value=np.zeros((1, n_actions)), validate_shape=True, trainable=True,
                            dtype=tf.float32, name='bias')
    output = tf.matmul(state_holder, theta_var) + bias_var

with tf.name_scope('policy'):
    action_probabilities = tf.nn.softmax(output, dim=-1, name='action_probabilities')
    testing_action_choice = tf.argmax(output, dimension=1, name='testing_action_choice')

with tf.name_scope('loss'):
    # -> n_batches x 1
    chosen_action_probability = tf.reduce_sum(tf.multiply(action_probabilities + 1e-8, actions_one_hot_holder), reduction_indices=1, name='chosen_action_probability')
    log_probability = tf.log(chosen_action_probability, name='log_probability')

    # # -> 1
    log_probability_sum = tf.reduce_sum(tf.log(chosen_action_probability), reduction_indices=0,
                                        name='log_probability_sum')

    # # -> 1
    discounted_rewards_sum = tf.reduce_sum(discounted_rewards_holder, reduction_indices=0,
                                           name='discounted_rewards_sum')

    # -> 1
    L_theta = -(log_probability_sum * (discounted_rewards_sum - constant_baseline))

with tf.name_scope('train'):
    gd_opt = tf.train.AdamOptimizer(learning_rate)  # .minimize(L_theta, global_step=global_step)
    grad_apply = gd_opt.minimize(L_theta)

sess = tf.Session()  # FOR NOW everything is symbolic, this object has to be called to compute each value of Q

sess.run(tf.global_variables_initializer())

observation = env.reset()
batch_rewards = []
states = []
action_one_hots = []

episode_rewards = []
episode_rewards_list = []
episode_steps_list = []

step = 0
episode_no = 0
while episode_no <= n_train_trials:
    # env.render()
    step += 1

    action_probability_values = sess.run(action_probabilities,
                                         feed_dict={state_holder: observation.reshape(1, n_obs)})
    action = np.random.choice(np.arange(n_actions), p=action_probability_values.ravel())

    action_arr = np.zeros(n_actions)
    action_arr[action] = 1.
    action_one_hots.append(action_arr)

    states.append(observation)

    observation, reward, done, info = env.step(action)
    if step >= 200:
        done = True

    batch_rewards.append(reward)
    episode_rewards.append(reward)

    if len(batch_rewards) > 0 and done:
        batch_reward_length = len(batch_rewards)

        gradients = sess.run(grad_apply,
                                      feed_dict={actions_one_hot_holder: action_one_hots, state_holder: states,
                                                 discounted_rewards_holder: batch_rewards})

        action_one_hots = []
        states = []
        batch_rewards = []

    if done:
        episode_no += 1

        episode_rewards_list.append(np.sum(episode_rewards))
        episode_steps_list.append(step)

        episode_rewards = []

        step = 0

        observation = env.reset()

        if episode_no % print_per_episode == 0:
            print("Average steps in last {} episodes before episode {}".format(print_per_episode, episode_no),
                  np.mean(episode_steps_list[(episode_no - print_per_episode):episode_no]), '+-',
                  np.std(episode_steps_list[(episode_no - print_per_episode):episode_no])
                  )

plt.figure()
plt.plot(range(len(episode_rewards_list)), episode_rewards_list)
plt.title("Training")

observation = env.reset()

episode_rewards_list = []
episode_rewards = []
episode_steps_list = []

step = 0
episode_no = 0

print("Testing")
while episode_no <= n_test_trials:
    env.render()
    step += 1

    test_action, = sess.run([testing_action_choice],
                            feed_dict={state_holder: observation.reshape(1, n_obs)})

    observation, reward, done, info = env.step(test_action[0])
    if step >= 200:
        done = True
    episode_rewards.append(reward)

    if done:
        episode_no += 1

        episode_rewards_list.append(np.sum(episode_rewards))
        episode_steps_list.append(step)

        episode_rewards = []
        step = 0
        observation = env.reset()

        if episode_no % print_per_episode == 0:
            print("Average steps in last {} episodes before episode {}".format(print_per_episode, episode_no),
                  np.mean(episode_steps_list[(episode_no - print_per_episode):episode_no]), '+-',
                  np.std(episode_steps_list[(episode_no - print_per_episode):episode_no])
                  )

plt.figure()
plt.plot(range(len(episode_rewards_list)), episode_rewards_list)
plt.title("Test")
plt.show()
