#!/usr/bin/env python
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt

import gym

__author__ = "Anand Subramoney"

def get_all_coords(I):
    I = I[35:195]  # crop
    I = I[::2, ::2, 0]  # downsample by factor of 2
    I[I == 144] = 0  # erase background (background type 1)
    I[I == 109] = 0  # erase background (background type 2)
    I[I != 0] = 1  # everything else (paddles, ball) just set to 1e
    opponent_slice = I[:, 9:10]
    # opponent_x = 10.
    opponent_y = ((np.argmax(opponent_slice, axis=0) + (
        opponent_slice.shape[0] - np.argmax(np.array(list(reversed(opponent_slice))), axis=0))) / 2)[0]
    self_slice = I[:, 70:71]
    self_x = 70.
    self_y = ((np.argmax(self_slice, axis=0) + (
        self_slice.shape[0] - np.argmax(np.array(list(reversed(self_slice))), axis=0))) / 2)[0]
    ball_slice = I[:, 10:70]
    ball_x = 10. + np.argmax(np.sum(ball_slice, axis=0))
    ball_y = np.argmax(np.sum(ball_slice, axis=1))
    return float(opponent_y - self_y), float(ball_x - self_x), float(ball_y - self_y)
    # return self_x, self_y, opponent_x, opponent_y, ball_x, ball_y


# downsampling
def prepro(I, I_prev):
    """ prepro 210x160x3 uint8 frame into 6400 (80x80) 1D float vector """
    opp_rel_y, ball_rel_x, ball_rel_y = get_all_coords(I)
    opp_rel_y_prev, ball_rel_x_prev, ball_rel_y_prev = get_all_coords(I_prev)
    return np.array([opp_rel_y, ball_rel_x, ball_rel_y, opp_rel_y_prev, ball_rel_x_prev, ball_rel_y_prev])


env = gym.make("Pong-v0")

n_actions = 2
action_list = [2, 3]
n_obs = 6

print_per_episode = 100

max_episode_len = 200
batch_size = 200
n_train_trials = 1000
n_test_trials = 100
gamma = 0.9999
learning_rate = 0.001

state_holder = tf.placeholder(dtype=tf.float32, shape=(None, n_obs), name='symbolic_state')
actions_one_hot_holder = tf.placeholder(dtype=tf.float32, shape=(None, n_actions),
                                        name='symbolic_actions_one_hot_holder')

discounted_rewards_holder = tf.placeholder(dtype=tf.float32, shape=None, name='symbolic_reward')


with tf.name_scope('linear'):
    theta_var = tf.Variable(initial_value=np.random.rand(n_obs, n_actions), validate_shape=True,
                            trainable=True, dtype=tf.float32, name='theta')
    bias_var = tf.Variable(initial_value=np.zeros((1, n_actions)), validate_shape=True, trainable=True,
                            dtype=tf.float32, name='bias')
    output = tf.matmul(state_holder, theta_var) + bias_var

with tf.name_scope('policy'):
    action_probabilities = tf.nn.softmax(output, dim=-1, name='action_probabilities')
    testing_action_choice = tf.argmax(output, dimension=1, name='testing_action_choice')

with tf.name_scope('loss'):
    # -> n_batches x 1
    chosen_action_probability = tf.reduce_sum(tf.multiply(action_probabilities + 1e-8, actions_one_hot_holder),
                                              reduction_indices=1,
                                              name='chosen_action_probability')
    log_probability = tf.log(chosen_action_probability, name='log_probability')

    # # -> 1
    log_probability_sum = tf.reduce_sum(tf.log(chosen_action_probability), reduction_indices=0,
                                        name='log_probability_sum')

    # # -> 1
    discounted_rewards_sum = tf.reduce_sum(discounted_rewards_holder, reduction_indices=0,
                                           name='discounted_rewards_sum')

    # -> 1
    L_theta = -(log_probability_sum * (discounted_rewards_sum - 1.))

with tf.name_scope('train'):
    global_step = tf.Variable(0, trainable=False)
    # gd_opt = tf.train.AdamOptimizer(learning_rate)
    gd_opt = tf.train.RMSPropOptimizer(learning_rate)
    grad_apply = gd_opt.minimize(L_theta, global_step=global_step)

sess = tf.Session()  # FOR NOW everything is symbolic, this object has to be called to compute each value of Q


sess.run(tf.global_variables_initializer())

prev_observation = observation = env.reset()
batch_rewards = []
states = []
action_one_hots = []

single_episode_rewards = []
episode_rewards_list = []
episode_steps_list = []

step = 0
episode_no = 0
while episode_no <= n_train_trials:
    # env.render()
    step += 1

    processed_observation = prepro(observation, prev_observation)

    action_probability_values = sess.run(action_probabilities,
                                         feed_dict={state_holder: processed_observation.reshape(1, n_obs)})
    action_idx = np.random.choice(range(n_actions), p=action_probability_values.ravel())
    action = action_list[action_idx]

    action_arr = np.zeros(n_actions)
    action_arr[action_idx] = 1.
    action_one_hots.append(action_arr)

    states.append(processed_observation)

    prev_observation = observation
    observation, reward, done, info = env.step(action)

    batch_rewards.append(reward)
    single_episode_rewards.append(reward)

    # if step % batch_size == 0 or (len(batch_rewards) > 0 and done):
    if reward != 0 or (len(batch_rewards) > 0 and done):
        batch_reward_length = len(batch_rewards)

        for i in range(batch_reward_length):
            batch_rewards[i] *= (gamma ** (batch_reward_length - i - 1))

        gradients = sess.run(grad_apply,
                                      feed_dict={actions_one_hot_holder: action_one_hots, state_holder: states,
                                                 discounted_rewards_holder: batch_rewards})

        action_one_hots = []
        states = []
        batch_rewards = []

    if done:
        episode_no += 1

        episode_rewards_list.append(np.sum(single_episode_rewards))
        episode_steps_list.append(step)

        single_episode_rewards = []

        step = 0

        observation = env.reset()

        if episode_no % print_per_episode == 0:
            print("Average REWARDS in last {} episodes before episode {}".format(print_per_episode, episode_no),
                  np.mean(episode_rewards_list[(episode_no - print_per_episode):episode_no]), '+-',
                  np.std(episode_rewards_list[(episode_no - print_per_episode):episode_no])
                  )
            print("Average STEPS in last {} episodes before episode {}".format(print_per_episode, episode_no),
                  np.mean(episode_steps_list[(episode_no - print_per_episode):episode_no]), '+-',
                  np.std(episode_steps_list[(episode_no - print_per_episode):episode_no])
                  )

plt.figure()
plt.plot(range(len(episode_rewards_list)), episode_rewards_list)
plt.title("Training")

env.monitor.start('monitor/pong-experiment-1')
prev_observation = observation = env.reset()

episode_rewards_list = []
single_episode_rewards = []
episode_steps_list = []

step = 0
episode_no = 0

print("Testing")

while episode_no <= n_test_trials:
    env.render()
    step += 1

    processed_observation = prepro(observation, prev_observation)
    test_action_idx, = sess.run([testing_action_choice],
                                feed_dict={state_holder: processed_observation.reshape(1, n_obs)})
    test_action = action_list[test_action_idx[0]]

    prev_observation = observation
    observation, reward, done, info = env.step(test_action)
    single_episode_rewards.append(reward)

    if done:
        episode_no += 1

        episode_rewards_list.append(np.sum(single_episode_rewards))
        episode_steps_list.append(step)

        single_episode_rewards = []
        step = 0
        observation = env.reset()

        if episode_no % print_per_episode == 0:
            print("Average REWARDS in last {} episodes before episode {}".format(print_per_episode, episode_no),
                  np.mean(episode_rewards_list[(episode_no - print_per_episode):episode_no]), '+-',
                  np.std(episode_rewards_list[(episode_no - print_per_episode):episode_no])
                  )
            print("Average STEPS in last {} episodes before episode {}".format(print_per_episode, episode_no),
                  np.mean(episode_steps_list[(episode_no - print_per_episode):episode_no]), '+-',
                  np.std(episode_steps_list[(episode_no - print_per_episode):episode_no])
                  )

env.monitor.close()

plt.figure()
plt.plot(range(len(episode_rewards_list)), episode_rewards_list)
plt.title("Test")
plt.show()
